from collections import Counter

import networkx as nx
import pandas as pd

from dhlab.api.nb_search_api import get_df, get_konks
from dhlab.legacy.nbtext import frame, frame_sort
from dhlab.text.nbtokenizer import tokenize


def phrase_plots(
    phrase_sets,
    title='aftenposten',
    fra=1960,
    til=2020,
    step=5,
    rot=0,
    colours=['r', 'b', 'g', 'y', 'm', 'c']
):
    df_all = []
    for f in phrase_sets:
        df_all.append(frame(get_df(f, title=title), ', '.join(f)))
    df = pd.concat(df_all, sort=False)
    df.index = df.index.astype(int)
    df = df.sort_index()
    df['bins'] = pd.cut(df.index, range(fra, til, step), precision=0)
    df.groupby('bins').sum().plot(
        kind='bar', color=colours, figsize=(
            15, 5), rot=rot)
    return


def phrase_plots_anno(
    phrase_sets,
    title='aftenposten',
    fra=1960,
    til=2020,
    rot=0,
    colours=['r', 'b', 'g']
):
    df_all = []
    for f in phrase_sets:
        df_all.append(frame(get_df(f, title=title), ', '.join(f)))
    df = pd.concat(df_all, sort=False)
    df.index = df.index.astype(int)
    df = df.sort_index()
    #df['bins'] = pd.cut(df.index, range(fra, til, step), precision=0)
    df.plot(kind='bar', figsize=(15, 5), rot=rot, color=colours)
    return


def graph_from_df(df, threshold=100):
    edges = []
    normalizer = {(x, y): df.stack()[(x, x)] *
                  df.stack()[(y, y)] for (x, y) in df.stack().index}
    for (x, y) in df.stack().index:
        if x != y:
            if df.stack()[(x, y)] > threshold:
                edges.append([x, y, df.stack()[(x, y)] / normalizer[(x, y)]])
    G = nx.Graph()
    G.add_weighted_edges_from(edges)
    return G


def get_all_konks(term, urns):
    konks = []
    for u in urns:
        konks += get_konks(u, term)
    return konks


def collocations_from_nb(word, corpus):
    """Get a concordance, and count the words in it.
    Assume konks reside a dataframe with columns 'after' and 'before'"""
    concordance = frame(get_all_konks(word, corpus))
    return frame_sort(frame(Counter(tokenize(
        ' '.join(concordance['after'].values + concordance['before'].values))), word))


def count_from_conc(concordance):
    """From a concordance, count the words in it.
    Assume konks reside a dataframe with columns 'after' and 'before'"""
    word = concordance['word'][0]
    return frame_sort(
        frame(
            Counter(
                tokenize(
                    ' '.join(
                        concordance['after'].values +
                        concordance['before'].values))
            ),
            word
        )
    )



from pandas import DataFrame
import pandas as pd

from dhlab.api.dhlab_api import document_corpus, get_metadata, evaluate_documents
from typing import Union

class Corpus:
    def __init__(
            self,
            doctype=None,
            author=None,
            freetext=None,
            fulltext=None,
            from_year=None,
            to_year=None,
            from_timestamp=None,
            to_timestamp=None,
            title=None,
            ddk=None,
            subject=None,
            lang=None,
            limit=10):

        self.corpus = document_corpus(
            doctype,
            author,
            freetext,
            fulltext,
            from_year,
            to_year,
            from_timestamp,
            to_timestamp,
            title,
            ddk,
            subject,
            lang,
            limit
        )

        self.size = len(self.corpus)

    def __repr__(self) -> str:
        """
        Return the string representation of the corpus datafrane
        """
        return self.corpus.__repr__()
    
    def _repr_html_(self) -> Union[str, None]:
        """
        Return the HTML representation of the corpus datafrane
        """
        return self.corpus._repr_html_()
    

    def add(self, corpus = None):
        """Add a corpus to existing corpus"""
        self.corpus = pd.concat([self.corpus, corpus.corpus]).drop_duplicates().reset_index(drop=True)
        self.size = len(self.corpus)
    
    def extend_from_identifiers(self, identifiers=None):
        corpus = get_metadata(urnlist(identifiers))
        self.corpus = pd.concat([self.corpus, corpus]).drop_duplicates().reset_index(drop=True)
        self.size = len(self.corpus)
        
    def evaluate_words(self, wordbags = None):
        df = evaluate_documents(wordbags = wordbags, urns = list(self.corpus.urn))
        df.index = df.index.astype(int)
        cols = df.columns
        df = pd.concat([df, self.corpus[['dhlabid','urn']].set_index('dhlabid')], axis = 1)
        df = df.set_index('urn')
        return df[cols].fillna(0)
    
class EmptyCorpus(Corpus):
    def __init__(self):
        self.corpus = pd.DataFrame()
        self.size = 0
        
class Corpus_from_identifiers(Corpus):
    def __init__(self, identifiers=None):
        self.corpus = get_metadata(urnlist(identifiers))
        self.size = len(self.corpus)

def urnlist(corpus):
    """Try to pull out a list of URNs from corpus"""

    if isinstance(corpus, Corpus):
        urnlist = list(corpus.corpus.urn)
    elif isinstance(corpus, DataFrame):
        urnlist = list(corpus.urn)
    elif isinstance(corpus, list):
        urnlist = corpus
    else:
        urnlist = []
    return urnlist



from collections import Counter

import pandas as pd
from bs4 import BeautifulSoup

from dhlab.legacy.nbtext import gather_wordlists
from dhlab.text.nbtokenizer import tokenize


def text_from_html_file(filename):
    """Get a flat list of tokens from a file"""
    with open(filename, encoding='utf-8') as f:
        content = f.read()
    soup = BeautifulSoup(content, 'html.parser')
    text = []
    for para in soup.find_all('p'):
        text += tokenize(para.text)
    return text


def growth_diagram_from_text(tekst, ordlister, window=5000, pr=100):
    """Serier fra ordlistene sammen.

    ordlister er en dictionary eller en liste av en liste over ord.
    """
    rammer = {}
    c = gather_wordlists(ordlister)

    for key in c:
        rammer[key] = []
    for i in range(0, len(tekst), pr):
        # count words of size windows - check below if text is overrun
        word_counts = Counter(tekst[i: i + window])
        for key in c:
            key_counts = sum([word_counts[word] for word in c[key]])
            rammer[key].append(key_counts)
        if i + window > len(tekst):
            break
    return pd.DataFrame(rammer)



import pytest
from dhlab import nbtext


@pytest.mark.parametrize(
    "input_data",
    [
        "https://www.nb.no/items/URN:NBN:no-nb_digibok_2019100726008",
        "URN:NBN:no-nb_digibok_2019100726008",
        "digibok_2019100726008",
        ["2019100726008", "2019100726008", "2019100726008"]
    ]
)
def test_pure_urn(input_data):
    # when
    result = set(nbtext.pure_urn(input_data))
    # then
    assert len(result) == 1
    assert result == {"2019100726008"}



import pandas as pd

from dhlab.api.dhlab_api import get_chunks, get_chunks_para


class Chunks:
    """Create chunks from a text."""

    def __init__(self, urn=None, chunks=1000):
        """
        :param urn: str or list
        :param chunks: {'para', 'avsn'} or int
        """
        # switch function based on value of chunks - a number or a string
        # indicating whether each paragraph is a chunk, else chop up the text
        # in pieces according to chunks

        if isinstance(chunks, str):
            if chunks.startswith("para") or chunks.startswith("avsn"):
                self.chunks = get_chunks_para(urn=urn)
            else:
                self.chunks = {}
        else:
            self.chunks = get_chunks(urn=urn, chunk_size=chunks)

    def to_pandas(self):
        """Vectorize into a pandas dataframe with words a index"""
        return pd.DataFrame(self.chunks).transpose()



from datetime import datetime

from dhlab.api.dhlab_api import ngram_book, ngram_news
from dhlab.ngram.nb_ngram import nb_ngram


class Ngram:
    """Top level class for ngrams"""

    def __init__(self, words=None, from_year=None, to_year=None, doctype=None, lang='nob') -> None:

        self.date = datetime.now()
        if to_year is None:
            to_year = self.date.year
        if from_year is None:
            from_year = 1950

        self.from_year = from_year
        self.to_year = to_year
        self.words = words
        self.lang = lang
        if not doctype is None:
            if 'bok' in doctype:
                doctype = 'bok'
            elif 'avis' in doctype:
                doctype = 'avis'
            else:
                doctype = 'bok'
        else:
            doctype = 'bok'
        ngrm = nb_ngram(terms=', '.join(words), corpus=doctype, years=(from_year, to_year), lang = lang)
        ngrm.index = ngrm.index.astype(str)
        self.ngram = ngrm

    def plot(self, **kwargs):
        self.ngram.plot(**kwargs)

    def compare(self, another_ngram):
        start_year = max(datetime(self.from_year, 1, 1),
                         datetime(another_ngram.from_year, 1, 1)).year
        end_year = min(datetime(self.to_year, 1, 1), datetime(another_ngram.to_year, 1, 1)).year
        transposed_ngram = self.ngram.loc[str(start_year):str(end_year)].transpose()
        sum_other_ngram = another_ngram.ngram[str(start_year):str(end_year)].transpose().sum()
        compare = (transposed_ngram / sum_other_ngram).transpose()
        return compare


class NgramBook(Ngram):
    """Extract ngrams using metadata with functions to be inherited."""

    def __init__(
            self,
            words=None,
            title=None,
            publisher=None,
            city=None,
            lang='nob',
            from_year=None,
            to_year=None,
            ddk=None,
            subject=None
    ):

        super().__init__(words, from_year, to_year, lang)
        self.date = datetime.now()
        if to_year is None:
            to_year = self.date.year
        if from_year is None:
            from_year = 1950
        self.from_year = from_year
        self.to_year = to_year
        self.words = words
        self.title = title
        self.publisher = publisher
        self.city = city
        self.lang = lang
        self.ddk = ddk
        self.subject = subject
        self.ngram = ngram_book(word=words, title=title, publisher=publisher, lang=lang, city=city,
                                period=(from_year, to_year), ddk=ddk, topic=subject)
        # self.cohort =  (self.ngram.transpose()/self.ngram.transpose().sum()).transpose()
        return None


class NgramNews(Ngram):
    def __init__(
            self,
            words=None,
            title=None,
            city=None,
            from_year=None,
            to_year=None
    ):
        self.date = datetime.now()
        if to_year is None:
            to_year = self.date.year
        if from_year is None:
            from_year = 1950
        self.from_year = from_year
        self.to_year = to_year
        self.words = words
        self.title = title
        self.ngram = ngram_news(word=words, title=title, period=(from_year, to_year))
        # self.cohort =  (self.ngram.transpose()/self.ngram.transpose().sum()).transpose()



import pandas as pd

from dhlab.api.nb_ngram_api import get_ngram


def nb_ngram(terms: str,
             corpus: str = 'bok',
             smooth: int = 3,
             years: tuple = (1810, 2010),
             mode: str = 'relative',
            lang: str = 'nob'):
    """Extract N-gram frequencies from given ``terms`` and ``years``.

    :param terms: comma
    :param corpus:
    :param smooth:
    :param years:
    :param mode:
    :param lang:
    :return: A sorted Pandas DataFrame index

    :meta private:
    """
    df = ngram_conv(get_ngram(terms, corpus=corpus, lang = lang), smooth=smooth, years=years, mode=mode)
    df.index = df.index.astype(int)
    return df.sort_index()

## tar tilbake til original den her virker ikke LGJ
def ngram_conv_old(ngrams, smooth: int = 1, years: tuple = (1810, 2013), mode: str = 'relative'):
    """Construct a dataframe with ngram mean frequencies per year over a given time period.

    :param ngrams: TODO: FIll in appropriate type and description.
    :param smooth: Smoothing factor for the graph visualisation.
    :param years: Tuple with start and end years for the time period of interest
    :param mode: Frequency measure. Defaults to 'relative'.
    :return: pandas dataframe with mean values for each year

    :meta private:
    """
    ngc = {}
    # check if relative frequency or absolute frequency is in question
    if mode.startswith('rel') or mode == 'y':
        arg = 'y'
    else:
        arg = 'f'
    for x in ngrams:
        if x and isinstance(x, list):
            ngc[x['key']] = {
                z['x']: z[arg]
                for z in x['values']
                if years[1] >= int(z['x']) >= years[0]
            }
    return pd.DataFrame(ngc).rolling(window=smooth, win_type='triang').mean()


def ngram_conv(ngrams, smooth=1, years=(1810,2013), mode='relative'):
    ngc = {}
    # check if relative frequency or absolute frequency is in question
    if mode.startswith('rel') or mode=='y':
        arg = 'y'
    else:
        arg = 'f'
    for x in ngrams:
        if x != []:
            ngc[x['key']] = {z['x']:z[arg] for z in x['values'] if int(z['x']) <= int(years[1]) and int(z['x']) >= int(years[0])}
    return pd.DataFrame(ngc).rolling(window=smooth, win_type='triang').mean()


import pandas as pd
import requests
from typing import List

from dhlab.constants import BASE_URL


def get_metadata(urns: List = None):
    """Fetch metadata from a list of urns.

    :param list urns: uniform resource names, example: ``["URN:NBN:no-nb_digibok_2011051112001", ...]``
    """
    params = locals()
    r = requests.post(f"{BASE_URL}/get_metadata", json=params)
    return pd.DataFrame(r.json())




import re

import pandas as pd

from dhlab.api.dhlab_api import get_document_frequencies, concordance, urn_collocation
from dhlab.text.corpus import urnlist

from typing import  Union


# convert cell to a link
def make_link(row):
    r = "<a target='_blank' href = 'https://urn.nb.no/{x}'>{x}</a>".format(
        x=str(row))
    return r


# find hits a cell
def find_hits(x): return ' '.join(re.findall("<b>(.+?)</b", x))


class Concordance():
    """Wrapper for concordance function with added functionality"""

    def __init__(self, corpus=None, query=None, window=20, limit=500):

        self.concordance = concordance(urns=urnlist(corpus), words=query, window=window, limit=limit)
        self.concordance['link'] = self.concordance.urn.apply(make_link)
        self.concordance = self.concordance[['link', 'urn', 'conc']]
        self.concordance.columns = ['link', 'urn', 'concordance']
        self.corpus = corpus
        self.size = len(self.concordance)

    def __repr__(self) -> str:
        """
        Return the string representation of the concordance datafrane
        """
        return self.corpus.__repr__()
    
    def _repr_html_(self) -> Union[str, None]:
        """
        Return the HTML representation of the concordance datafrane
        """
        return self.corpus._repr_html_()   

    def show(self, n=10, style=True):
        if style:
            result = self.concordance.sample(min(n, self.size))[
                ['link', 'concordance']].style
        else:
            result = self.concordance.sample(min(n, self.size))
        return result


class Collocations():
    """Collocations """

    def __init__(
        self,
        corpus=None,
        words=None,
        before=10,
        after=10,
        reference=None,
        samplesize=20000,
        alpha=False,
        ignore_caps=False
    ):
        """Create collocations object

        :param corpus: target corpus, defaults to None
        :type corpus: dh.Corpus, optional
        :param words: target words(s), defaults to None
        :type words: str or list, optional
        :param before: words to include before, defaults to 10
        :type before: int, optional
        :param after: words to include after, defaults to 10
        :type after: int, optional
        :param reference: reference frequency list, defaults to None
        :type reference: pd.DataFrame, optional
        :param samplesize: _description_, defaults to 20000
        :type samplesize: int, optional
        :param alpha: Only include alphabetical tokens, defaults to False
        :type alpha: bool, optional
        :param ignore_caps: Ignore capitalized letters, defaults to False
        :type ignore_caps: bool, optional
        """
        if isinstance(words, str):
            words = [words]
        
        
                
        coll = pd.concat(
            [
                urn_collocation(
                    urns=urnlist(corpus),
                    word=w,
                    before=before,
                    after=after,
                    samplesize=samplesize
                )
                for w in words
            ]
        )[['counts']]
        
        if alpha:
            coll = coll.loc[[x for x in coll.index if x.isalpha()]]
            if reference is not None:
                reference = reference.loc[[x for x in reference.index if x.isalpha()]]
            
            
        if ignore_caps:
            coll.index = [x.lower() for x in coll.index]
            if reference is not None:
                reference.index = [x.lower() for x in reference.index]

        self.coll = coll.groupby(coll.index).sum()
        self.reference = reference
        self.before = before
        self.after = after
       

        if reference is not None:
            teller = self.coll.counts / self.coll.counts.sum()
            divisor = self.reference.iloc[:, 0] / self.reference.iloc[:, 0].sum()
            self.coll['relevance'] = teller / divisor
            
    def __repr__(self) -> str:
        """
        Return the string representation of the collocation datafrane
        """
        return self.coll.__repr__()
    
    def _repr_html_(self) -> Union[str, None]:
        """
        Return the HTML representation of the collocation datafrane
        """
        return self.coll._repr_html_()       

    def show(self, sortby='counts', n=20):
        return self.coll.sort_values(by=sortby, ascending=False)

    def keywordlist(self, top=200, counts=5, relevance=10):
        mask = self.coll[self.coll.counts > counts]
        mask = mask[mask.relevance > relevance]
        return list(mask.sort_values(
            by='counts', ascending=False).head(200).index)


class Counts():
    """Provide counts for a corpus - shouldn't be too large"""

    def __init__(self, corpus=None, words=None):
        if corpus is None and words is None:
            self.counts = None
        elif not corpus is None:
            # count - if words is none result will be as if counting all words
            # in the corpus
            self.counts = get_document_frequencies(
                urns=urnlist(corpus), cutoff=0, words=words)


    def __repr__(self) -> str:
        """
        Return the string representation of the counts datafrane
        """
        return self.counts.__repr__()
    
    def _repr_html_(self) -> Union[str, None]:
        """
        Return the HTML representation of the counts datafrane
        """
        return self.counts._repr_html_()  
        


import pandas as pd

from dhlab.api.dhlab_api import get_places, geo_lookup, ner_from_urn
from dhlab.text.parse import NER, Models

spacy = Models()

class GeoData:
    """Fetch place data from a text (book, newspaper or ...) identified by URN
    with an appropriate and available spacy model. The models are gotten at by instantiaiting class Models()"""
    
    def __init__(self, urn=None, model=spacy.models[0]):
        df = NER(urn = urn, model = model).ner
        self.place_names = df[df['ner'].str.contains('LOC')]
        
    def add_geo_locations(self, feature_class = None, feature_code = None):
        """Get location data for the names in object, attribute self.place_names"""
        chunksize = 900
        names = list(self.place_names.token)
        length = len(names)
        
        # GeoNames takes 1000 names at a time so chunk things up
        # Each GeoNames object has an attribute .places a pandas dataframe
        
        chunks = [
            GeoNames(
                names[i:i+chunksize], 
                feature_class = feature_class, 
                feature_code = feature_code
            ).places for i in range(0, length, chunksize)
        ] 
        
        self.places = pd.concat(chunks) 

        
class GeoNames:
    """Fetch data from a list of names"""
    def __init__(self, names, feature_class = None, feature_code = None):
        self.places = geo_lookup(names, feature_class = feature_class, feature_code = feature_code)
        
        


from urllib.parse import urlparse

import requests
from IPython.display import HTML


def code_toggle(button_text = "Klikk for å vise/skjule kodeceller"):
    from IPython.display import HTML, display

    display(
        HTML(
        '''<div>
                <style>
                 .mybutton {
                    background-color: lightgrey;
                    border: none;
                    color: white;
                    padding: 10px 16px;
                    text-align: center;
                    text-decoration: none;
                    display: inline-block;
                    font-size: 16px;
                    margin: 4px 2px;
                    cursor: pointer;
                }
            </style>
            <script>
                code_show=true; 
                function code_toggle() {
                 if (code_show){
                 $('div.input').hide();
                 } else {
                 $('div.input').show();
                 }
                 code_show = !code_show
                } 
                $( document ).ready(code_toggle);
            </script>
            <form  action="javascript:code_toggle()">
                <input class='mybutton' type="submit" value=''' + '"'  + button_text + '"' + '''>
            </form>
        </div>'''
        ))


def css(url = "https://raw.githubusercontent.com/Yoonsen/Modules/master/css_style_sheets/nb_notebook.css"):
    """Associate a css stylesheet with the notebook, 
    just specify a file or web reference, default is a custom css"""
    
    uri = urlparse(url)
    css_file = ""
    
    if uri.scheme.startswith('http'):
        query = requests.get(url)
        if query.status_code == 200:
            css_file  = query.text
    
    elif uri.scheme == "file": 
        # assume on form "file:/// on windows there is drive letter on unix not"
        file_path = url[7:]
        if file_path[2] == ':': # then windows drive reference
            file_path = file_path[1:]
        with open(file_path, encoding='utf-8') as file:
            css_file = file.read()
    else: 
        # assume string is a file locator
        with open(url, encoding='utf-8') as file:
            css_file = file.read()
    
    return HTML("<style>{css_code}</style>".format(css_code = css_file))



"""Constant variables used across modules."""

BASE_URL = "https://api.nb.no/dhlab"
NGRAM_API = "https://api.nb.no/dhlab/nb_ngram/ngram/query"
GALAXY_API = "https://api.nb.no/dhlab/nb_ngram_galaxies/galaxies/query"
